from pmtw.toolbox import Toolbox
from pmtw.settings import ToolboxSettings
from pmtw.usernotes import ToolboxNote, ToolboxUsernotes
from pmtw.classic import Note, Settings, Usernotes
from pmtw.puni import puni_Note, puni_UserNotes



import base64
import copy
import json
import re
import time as t
import zlib
from datetime import datetime

from prawcore.exceptions import NotFound

from pmtw.constants import (DEFAULT_IDENTIFIER, MAX_WIKI_SIZE, USERNOTES_PAGE,
                            USERNOTES_VERSION)
from pmtw.stream import revisions_stream


class ToolboxNote:
	"""Represents a single Toolbox Usernote."""
	def __init__(self, user, note, warning=None, time=None, mod=None, url=None, link='', notes=None):
		"""
		Construtor for the ToolboxNote class.

		Parameters
		----------
		user: String, praw.redditor object
			User the note is being left for
		note: String
			the text of the usernote
		warning: String
			the kind of warning the ntoe should be
		time: Integer
			a unix timestamp in seconds. Optional, will set to current time if
			not specified.
		mod: String, praw.redditor
			the moderator issuing the note. Optional, if not set will be set to
			the user of the praw instance.
		url: String
			url attached to the note. If present, gets converted to a toolbox
			link as `self.link`
		link: String
			the toolbox style short link. if set, will expand to a full url
		notes: ToolboxUsernotes
			if set to a ToolboxUsernotes instance, immediately adds the note.

		"""
		self.user = str(user)
		self.note = note
		self.warning = warning
		self.time = time if time else int(t.time())
		self.human_time = datetime.fromtimestamp(self.time).__str__()
		self.mod = str(mod)
		if link != '': self.link = link
		else: self.link = self.__compress_url(url)
		if url: self.url = url
		else: self.url = self.__expand_link(link)

		if notes != None: notes.add(self)

	def __repr__(self):
		"""praw-style representation of the note."""
		return f"ToolboxNote(user='{self.user}', note='{self.note}', human_time='{self.human_time}')"

	def __dict__(self):
		"""represnts a dict of the note in the format it's stored uncompressed in the json"""
		return {'t': self.time, 'n': self.note, 'm': self.mod, 'w': self.warning, 'l': self.link}

	@staticmethod
	def __expand_link(link):
		"""
		expand toolbox's shortlinks to full urls
		Parameters
		----------
		link: String
			toolbox formatted link to expand to a url
		Returns
		-------
		String
			an expanded link

		"""
		if link == '': return None #if no link, nothing to do
		parts = link.split(',') # 'l,abcde,fghij' format for comments, 'l,kmlno' for links
		if parts[0] == "m": return f'https://reddit.com/message/messages/{parts[1]}'
		if len(parts) > 2: return f'https://reddit.com/comments/{parts[1]}/-/{parts[2]}'
		elif len(parts) > 1: return f'https://reddit.com/{parts[1]}'
		else: return link #modmail links are stored as full links

	@staticmethod
	def __compress_url(url):
		"""
		compress a url to toolbox's `l,` format, if applicable

		Parameters
		----------
			url: String
				url to compress

		Returns
		-------
		String
			the toolbox-formatted link. Recognizes a wide variety of reddit
			links. The following is a non-exhaustive list to give examples:
			https://www.reddit.com/r/subreddit/abcdef/comments/ghijkl/
			http://old.reddit.com/r/subreddit/abcdef/
			http://reddit.com/r/subreddit/abcdef
			http://reddit.com/abcdef
			https://redd.it/abcdef
			reddit.com/abcdef

		"""
		if not url: return '' # If link is empty, nothing to compress
		if url.startswith("l,") or url.startswith("m,"): return url # Link already compressed
		if "mod.reddit" in url: return url # if link is to modmail message, nothing to compress
		match_re = re.compile(r'(https?:\/\/([a-z]{3}\.)?)?redd\.?it(\.com)?\/(r\/[a-z]*\/)?(comments\/)?', re.IGNORECASE)
		if not re.match(match_re, url): raise ValueError(f"cannot format {url}")
		url = re.sub(match_re,"", url)
		url = url.rstrip("/").split("/")
		if len(url) == 3: return f"l,{url[0]},{url[2]}" #return comment link format "l,abcde,fghij"
		else: return f"l,{url[0]}" # Return post link format "l,abcde"


class ToolboxUsernotes:
	"""Represents the Toolbox Usernotes page."""
	
	def __init__(self, subreddit, identifier=DEFAULT_IDENTIFIER, settingsWarnings=[], lazy=False):
		"""
		Construtor for the ToolboxUsernotes class.

		Parameters
		----------
		subreddit: praw.Subreddit object
			The praw object representing the subreddit to pull notes from
		lazy: Bool
			If False, fetch notes from Reddit on instantiation
		identifier: String
			string to identify all actions taken in the wiki save description
		settingsWarnings: List
			Note types which are in Settings which may or may not be present in 
			the usernotes wiki page
		"""
		self.warnings = []
		self.__subreddit = subreddit
		self.__identifier = identifier
		self.__usernotesJSON = {}
		self.__settingsWarnings = settingsWarnings

		if not lazy: self.load()

	def __str__(self):
		"""return the uncompessed json from the usernotes page as a string"""
		return str(self.__usernotesJSON)

	def __repr__(self):
		"""Set display for a ToolboxUsernotes object"""
		return f"ToolboxUsernotes(subreddit='{self.__subreddit}')"

	def __expand_json(self, notes):
		"""
		Private method used to expand the blob portion of usernotes
		
		Parameters
		----------
		notes: Dictionary
			dictionary representing the usernotes JSON

		Returns
		-------
		Dictionary
			dictionary representing the usernotes JSON with blob expanded


		"""
		json_blob = zlib.decompress(base64.b64decode(notes['blob'])).decode('utf-8') # decompress the Blob
		notes.pop('blob', None) # remove the Blob section from the json data	
		notes['users'] = json.loads(json_blob) # add decoded users section to json dictionary
		self.__usernotesJSON = notes
		return notes

	def __compress_json(self, notes=None):
		"""
		Private method to compress json usernotes into a blob for the wiki page
		
		Parameters
		----------
		notes: Dictionary
			usernotes wiki page in dictionary format. Works on self.__usernotesJSON
			if an explicit object is not supplied. This variable exists largely
			for backwards-compatibilty reasons.

		Returns
		-------
		String
			the compressed JSON object, per toolbox's compression methods
		
		"""
		if notes == None: notes = self.__usernotesJSON
		compressed_json = copy.copy(notes) #create a shallow copy of notes
		compressed_json.pop('users', None) #remove the users section from the json data
		blob_data = zlib.compress(json.dumps(notes['users']).encode('utf-8'), 9) #compress blob with zlib
		blob = base64.b64encode(blob_data).decode('utf-8') #encode the blob in base64
		compressed_json['blob'] = blob #insert the blob back into the json
		return json.dumps(compressed_json)

	def __get_warnings(self):
		""""
		Private method. Gets warnings from the usernotes json, and appends
		other available warnings passed to the constructor
		"""
		self.warnings = self.__usernotesJSON['constants']['warnings'].copy()
		if None not in self.warnings: self.warnings.append(None)
		for usernoteColor in self.__settingsWarnings:
			if usernoteColor not in self.warnings:
				self.warnings.append(usernoteColor)
		
	def __get_mod_index(self, mod):
		"""
		Private method. Toolbox stores the moderator as an index for a given 
		note, not the actual name. This method fetches the appropriate unstable 
		index for a given mod, adding the mod to the list of mods if needed.

		Parameters
		----------
		mod: String
			The mod to get the list index for

		Returns
		-------
		Integer
			the integer index of the mod

		"""
		try:
			mod_index = self.__usernotesJSON['constants']['users'].index(mod)
			return mod_index
		except ValueError:
			self.__usernotesJSON['constants']['users'].append(mod)
			return self.__usernotesJSON['constants']['users'].index(mod)

	def __get_warning_index(self, warning):
		"""
		Private method. Toolbox stores warnings as an index for a given 
		note, not the actual warning name. This method fetches the appropriate 
		unstable index for a given warning, adding the warning to the list of 
		warnings in the usernote page if the warning exists in settings but not
		in usernotes.

		Parameters
		----------
		warning: String
			The warning to get the list index for

		Returns
		-------
		Integer
			the integer index of the warning

		Raises
		------
		ValueError
			If the input warning is not valid for these usernotes

		"""
		try:
			warning_index = self.__usernotesJSON['constants']['warnings'].index(warning)
			return warning_index
		except ValueError:
			if warning in self.warnings:
				self.__usernotesJSON['constants']['warnings'].append(warning)
				return self.__usernotesJSON['constants']['warnings'].index(warning)
			else:
				raise ValueError(f"{warning} is not a valid warning.")

	def __match_username(self, username):
		"""
		Try to match a user with notes and get the proper capitalization for key
		""" 
		search_user = ""
		for user in self.list_users():
			if username.lower() == user.lower():
				search_user = user
				break
		if search_user == "":
			search_user = username
		return search_user

	def save(self, reason='Usernote update'):
		"""
		Save usernotes to Reddit, with whatever reason is specified. Will raise 
		an OverflowError if the current data is too big for the wiki page.

		Parameters
		----------
			reason: String
				Optional, set a custom reason for the wiki page description

		Returns
		-------
		String
			Wiki page update description

		Raises
		------
		OverflowError
			If he text is larger than the allowed 1mb wikipage size
		"""
		reason = f"{reason} {self.__identifier}"
		wikipage_data = self.__compress_json(self.__usernotesJSON)
		if len(wikipage_data) > MAX_WIKI_SIZE:
			raise OverflowError(f'Usernote data {len(wikipage_data) - MAX_WIKI_SIZE} bytes too big to insert')
		self.__subreddit.wiki[USERNOTES_PAGE].edit(content=wikipage_data, reason=reason)
		return reason

	def load(self):
		"""
		Fetch usernotes from Reddit. Initializes a wiki page if the page doesn't 
		currently exist.

		Returns
		-------
		String
			message informing of successful usernote load

		Raises
		------
		RuntimeError
			if the schema doesn't match the expected version.

		"""
		try:
			usernotes = self.__subreddit.wiki[USERNOTES_PAGE].content_md
			notes = json.loads(usernotes)
		except NotFound:
			initialJson = {"ver":USERNOTES_VERSION,"constants":{"users":[],"warnings":self.warnings}, "blob":""}
			self.sub.wiki.create(name=USERNOTES_PAGE, content=initialJson)
			self.sub.wiki[USERNOTES_PAGE].mod.update(listed=False, permlevel=2)
			self.__expand_json(initialJson)
		if notes['ver'] != USERNOTES_VERSION:
			raise RuntimeError(f"Usernotes Schema mismatch. PMTAW requires {USERNOTES_VERSION}, wiki page is {notes['ver']}")
		else:
			self.__expand_json(notes)
			self.__get_warnings()
		return "Usernotes loaded"

	def add(self, note, lazy=False):
		"""
		takes a ToolboxNote object and adds it on Reddit.

		Parameters
		----------
		note: ToolboxNote object
			note to add to the wiki
		timestamp: Int
			if set, delete note with the specified timestamp.
			if not set, delete all usernotes for a given user
		lazy: Bool
			If set to False, will immediately update the wiki page. if set to 
			True, will only modify the local usernote copy for manual saving later.

		Returns
		-------
		String
			string mirroring the description for the wiki update
		Raises
		------
		ValueError
			if the warning specified in the note does not exist in available 
			warning types for the configured subreddit.
		"""
		if lazy == False: self.load()
		
		new_note = note.__dict__()
		if new_note['m'] == 'None':
			new_note['m'] = self.__subreddit._reddit.user.me().name
		if new_note['w'] not in self.warnings:
				raise ValueError(f"{new_note['w']} is not a valid warning type")

		new_note['m'] = self.__get_mod_index(new_note['m'])
		if new_note['w'] == 'None': new_note['w'] = self.__get_warning_index(None)
		else: new_note['w'] = self.__get_warning_index(new_note['w'])

		user = self.__match_username(note.user)
		try:
			self.__usernotesJSON['users'][user]['ns'].insert(0, new_note)
			if lazy == False:
				self.save(f"create new note on user '{user}'")
				return f"create new note on user '{user}'"
		except KeyError:
			self.__usernotesJSON['users'][user] = {'ns': [new_note]}
			if lazy == False:
				self.save(f"create new note on user '{user}'")
				return f"create new note on user '{user}'"

	def remove(self, user, timestamp=-1, lazy=False):
		"""
		removes a specified user note, or all notes for a user. 

		Parameters
		----------
		user: String, praw.redditor
			user to remove notes for
		timestamp: Int
			Optional. if set, delete note with the specified timestamp.
			if not set, delete all usernotes for a given user
		lazy: Bool
			Optional. If set to False, will immediately update the wiki page. if
			set to True, will only modify the local usernote copy for manual 
			saving later.

		Returns
		-------
		String
			string mirroring what's put on the wiki

		Raises
		------
		KeyError
			if the note for the given timestamp isn't found

		"""
		user = str(user)
		user = self.__match_username(user)
		if lazy == False: self.load()

		if timestamp == -1:
			del self.__usernotesJSON['users'][user]
			if lazy == False: self.save(f"Deleted all notes on {user}")
			return f"Deleted all notes on {user}"
		else:
			notes_on_user = self.__usernotesJSON['users'][user]['ns']
			deleted = False
			for i in range(len(notes_on_user)):
				if notes_on_user[i]['t'] == timestamp:
					notes_on_user.pop(i)
					deleted = True
					break
			if deleted == False: raise KeyError(f"failed to find note timestamped {timestamp} for {user}")
			# Delete the user from the database if there are no notes left
			if len(notes_on_user) == 0:
				del self.__usernotesJSON['users'][user]
				if lazy == False: self.save(f"delete all notes on user '{user}'")
				return f"delete all notes on user '{user}'"
			else:
				if lazy == False: self.save(f"delete note {timestamp} on user '{user}'")
				return f"delete note {timestamp} on user '{user}'"

	def list_users(self, lazy=True):
		"""
		Returns a list of all users with notes.

		Parameters
		----------
		lazy: Bool
			if set to False, reload a fresh copy of usernotes before listing

		Returns
		-------
		List
			list of strings of every user with usernotes
		"""
		if lazy == False: self.load()
		return list(self.__usernotesJSON['users'].keys())

	def list_notes(self, user, lazy=True, reverse=False):
		"""
		Returns a list of all notes for the specified user. Case-insensitive

		Parameters
		----------
		user: String, praw.redditor
			user to fetch notes for
		lazy: Bool
			if set to False, reload a fresh copy of usernotes before listing
		reverse: Bool
			if set to True, display newest notes first

		Returns
		-------
		List
			list containing ToolboxUsernotes

		Raises
		------
		KeyError
			if user doesn't exist in usernotes

		"""	
		if lazy == False: self.load()
		user = str(user) #in case a praw user object is passed

		user = self.__match_username(user)

		try:
			users_notes = []
			for note in self.__usernotesJSON['users'][user]['ns']:
				users_notes.append(
					ToolboxNote(
						user = user,
						note = note['n'],
						time = note['t'],
						mod = self.__usernotesJSON['constants']['users'][note['m']],
						warning = self.__usernotesJSON['constants']['warnings'][note['w']],
						link = note['l']
					)
				)
			users_notes = sorted(users_notes, key = lambda x: x.time, reverse=reverse)
			return users_notes
		except KeyError: return []

	def list_all_notes(self, lazy=True, reverse=False):
		"""
		Returns a list of all notes in the wiki

		Parameters
		----------
		lazy: Bool
			if set to True, reload a fresh copy of usernotes before listing
		reverse: Bool
			If True, notes are listed newest to oldest
		Returns
		-------
		List
			list containing ToolboxUsernotes
		"""	
		if lazy == False: self.load()
		notes = []
		for user in self.list_users():
			for note in self.list_notes(user):
				notes.append(note)
		return sorted(notes, key = lambda x: x.time, reverse=reverse)

	def stream(self, pause_after=None, skip_existing=False):
		"""
		Yields new WikiRevision objects as they become available

		Parameters
		----------
		pause_after: [Optional] Integer (Default: `None`)
			An integer representing the number of requests that result in no new 
			items before this function yields `None`, effectively introducing a 
			pause into the stream. A negative value yields `None` after items 
			from a single response have been yielded, regardless of number of new 
			items obtained in that response. A value of `0` yields `None` after 
			every response resulting in no new items, and a value of `None` never 
			introduces a pause.
		skip_existing: [Optional] Boolean (Default: False)
			When `True`, this does not yield any results from the first request 
			thereby skipping any items that existed in the stream prior to starting 
			the stream.

		Yields
		------
		WikiRevision objects
		
		"""
		for revision in revisions_stream(
			sub=self.__subreddit,
			page=USERNOTES_PAGE,
			pause_after = pause_after,
			skip_existing = skip_existing
		):
			yield revision



import csv
import time as t

from pmtw.constants import DEFAULT_IDENTIFIER
from pmtw.settings import ToolboxSettings
from pmtw.usernotes import ToolboxUsernotes


class Toolbox:
	def __init__(self, subreddit, lazy=False, identifier=DEFAULT_IDENTIFIER):
		"""
		Constructor for the Toolbox class

		Parameters
		----------
		subreddit: praw.subreddit object
			the subreddit to use the Toolbox object with
		lazy: Bool
			Whether to fetch settings and usernotes on instantiation
		identifier: String
			string that get's appended to all wikiedit discriptions identifying
			pmtw as the actioner. Defaults to "via pmtw"
		"""
		self.settings = ''
		self.usernotes = ''
		self.identifier = identifier
		self.__subreddit = subreddit
		if not lazy: self._load()
			
	def __repr__(self):
		"""Set display for a Toolbox object"""
		return f"Toolbox(subreddit='{self.__subreddit}')"

	def _load(self):
		"""Load settings and Usernotes"""
		self.settings = ToolboxSettings(self.__subreddit, identifier=self.identifier)
		self.usernotes = ToolboxUsernotes(self.__subreddit, identifier=self.identifier, settingsWarnings=self.settings.warnings)

	def prune_notes(self, days=180, before=None, excludeKinds=[], dryRun=False):
		"""
		Bulk-removes old usernotes.

		Parameters
		----------
		days: Int
			number of days to prune notes back. Ignored if before is set.
		before: Int
			unix timestamp (in seconds) to remove notes before
		excludeKinds: List, String
			list of usernote warning types to preserve, even if older than the
			specified cutoff
		dryRun: Bool
			Simulates removal of notes without actually affecting them.
		
		Returns
		-------
		string
			notification of how many usernotes were deleted.

		"""
		self.usernotes.load()
		if isinstance(excludeKinds, str): excludeKinds = [excludeKinds]
		if before: pruneTime = before
		else: pruneTime = t.time() - (days * 86400)
		users_count = 0
		notes_count = 0
		preserved_count = 0
		for user in self.usernotes.list_users():
			users_notes = self.usernotes.list_notes(user)
			user_triggered = False
			for note in users_notes:
				if note.time > pruneTime: continue
				if note.warning in excludeKinds:
					preserved_count +=1
					continue
				notes_count += 1
				user_triggered = True
				self.remove(user, note.time, bulk_action=True)
			if user_triggered == True: users_count += 1
		if dryRun:
			self.usernotes.load()
			return f"prune would bulk delete {notes_count:,} notes on {users_count:,} users older than {days} days."
		else:
			self.usernotes.save(f"bulk deleted {notes_count:,} notes on {users_count:,} users older than {days} days")
			return f"bulk deleted {notes_count:,} notes on {users_count:,} users older than {days} days"

	def search_notes(self, query, kind="note", range="after", lazy=True):
		"""
		Search usernotes. 

		Parameters
		----------
		query: String
			the parameter to be searched
		kind: String
			one of: note, user, mod, warning, url, time
		range: String
			one of: before, after
		lazy: Bool
			if True, fetch from Reddit before searching

		Returns
		-------
		List
			a list of ToolboxUsernote objects matching the query parameter, or
			an empty list if there are no matches

		Raises
		------
		ValueError
			* when range is invalid if kind="time"
			* when kind is invalid
		"""
		if not lazy: self.usernotes.load()
		notes = self.usernotes.list_all_notes()
		results = []

		for note in notes:
			if kind == "time":
				if range == "after" and note.time > int(query): results.append(note)
				if range == "before" and note.time < int(query): results.append(note)
				else: raise ValueError(f"{range} is invalid. range options are ['before', 'after']")

			if kind == "note": search = note.note
			elif kind == "user": search = note.user
			elif kind == "mod": search = note.mod
			elif kind == "warning": search = note.warning
			elif kind == "url": search = note.url
			else: raise ValueError(f"{kind} is not a valid search option")

			search = str(search)
			search = search.lower()
			if query.lower() in search: results.append(note)
		return results

	def export_notes(
		self, 
		file="",
		fields=["user","note","warning","time","mod","link"],
		sortKey="time"
	):
		"""
		Exports Usernotes in a CSV file.

		Parameters
		----------
		file: String
			File name for the exported notes. Defaults to 
			"usernotes-{subreddit}-{current timestamp in seconds}.csv"
		fields: List:
			List of fields to export. Defaults to all fields
		sortKey: String
			Which key in fields to sort usernotes for in the exported csv file.
			defaults to time.

		Returns
		-------
		String
			Notification of success

		"""
		if file == "": file = f"usernotes-{self.__subreddit}-{int(t.time())}.csv"
		rows = []
		count = 0
		for note in self.usernotes.list_all_notes():
			row = []
			count += 1
			for field in fields: row.append(eval(f"note.{field}"))
			rows.append(row)
		try: rows = sorted(rows, key = lambda x: x[fields.index(sortKey)])
		except: rows = sorted(rows, key = lambda x: x[0])
		csv_dump = open(file, 'w')
		csv_writer = csv.writer(csv_dump)
		csv_writer.writerow(fields)
		for row in rows: csv_writer.writerow(row)
		csv_dump.close()
		return f"{count:,} usernotes exported to {file}"



import json
import urllib.parse
from dataclasses import dataclass
from typing import Any, List

from prawcore.exceptions import NotFound

from pmtw.constants import MAX_WIKI_SIZE, SETTINGS_PAGE, SETTINGS_VERSION, DEFAULT_IDENTIFIER
from pmtw.stream import revisions_stream

class JSONEncoder(json.JSONEncoder):
	# overload method default
	def default(self, obj):
		if isinstance(obj, ModMacro): return obj.to_dict()
		elif isinstance(obj, Reason): return obj.to_dict()
		elif isinstance(obj, RemovalReasons): return obj.to_dict()
		return obj.__dict__


@dataclass
class BanMacros:
	banNote: str
	banMessage: str

	@staticmethod
	def from_dict(obj: Any) -> 'BanMacros':
		_banNote = str(obj.get("banNote"))
		_banMessage = str(obj.get("banMessage"))
		return BanMacros(_banNote, _banMessage)


@dataclass
class DomainTag:
	name: str
	color: str

	@staticmethod
	def from_dict(obj: Any) -> 'DomainTag':
		_name = str(obj.get("name"))
		_color = str(obj.get("color"))
		return DomainTag(_name, _color)


@dataclass
class ModMacro:
	text: str
	title: str
	distinguish: bool
	ban: bool
	mute: bool
	remove: bool
	approve: bool
	lockthread: bool
	lockreply: bool
	sticky: bool
	archivemodmail: bool
	highlightmodmail: bool
	contextpost: bool
	contextcomment: bool
	contextmodmail: bool

	def to_dict(ModMacro):
		d = {
			"text":urllib.parse.quote(ModMacro.text),
			"title": ModMacro.title,
			"distinguish": ModMacro.distinguish,
			"ban": ModMacro.ban,
			"mute": ModMacro.mute,
			"remove": ModMacro.remove,
			"approve": ModMacro.approve,
			"lockthread": ModMacro.lockthread,
			"lockreply": ModMacro.lockreply,
			"sticky": ModMacro.sticky,
			"archivemodmail": ModMacro.archivemodmail,
			"highlightmodmail": ModMacro.highlightmodmail,
			"contextpost": ModMacro.contextpost,
			"contextcomment": ModMacro.contextcomment,
			"contextmodmail": ModMacro.contextmodmail
		}
		return d

	@staticmethod
	def from_dict(obj: Any) -> 'ModMacro':
		_text = urllib.parse.unquote(str(obj.get("text")))
		_title = str(obj.get("title"))
		_distinguish = bool(obj.get("distinguish"))
		_ban = bool(obj.get("ban"))
		_mute = bool(obj.get("mute"))
		_remove = bool(obj.get("remove"))
		_approve = bool(obj.get("approve"))
		_lockthread = bool(obj.get("lockthread"))
		_lockreply = bool(obj.get("lockreply"))
		_sticky = bool(obj.get("sticky"))
		_archivemodmail = bool(obj.get("archivemodmail"))
		_highlightmodmail = bool(obj.get("highlightmodmail"))
		_contextpost = bool(obj.get("contextpost"))
		_contextcomment = bool(obj.get("contextcomment"))
		_contextmodmail = bool(obj.get("contextmodmail"))
		return ModMacro(_text, _title, _distinguish, _ban, _mute, _remove, _approve, _lockthread, _lockreply, _sticky, _archivemodmail, _highlightmodmail, _contextpost, _contextcomment, _contextmodmail)


@dataclass
class Reason:
	text: str
	flairText: str
	flairCSS: str
	removePosts: bool
	removeComments: bool
	title: str
	flairTemplateID: str

	def to_dict(Reason):
		d = {
			"text": urllib.parse.quote(Reason.text),
			"flairText": Reason.flairText,
			"flairCSS": Reason.flairCSS,
			"removePosts": Reason.removePosts,
			"removeComments": Reason.removeComments,
			"title": Reason.title,
			"flairTemplateID": Reason.flairTemplateID
		}
		return d

	@staticmethod
	def from_dict(obj: Any) -> 'Reason':
		_text = urllib.parse.unquote(str(obj.get("text")))
		_flairText = str(obj.get("flairText"))
		_flairCSS = str(obj.get("flairCSS"))
		_removePosts = bool(obj.get("removePosts"))
		_removeComments = bool(obj.get("removeComments"))
		_title = str(obj.get("title"))
		_flairTemplateID = str(obj.get("flairTemplateID"))
		return Reason(_text, _flairText, _flairCSS, _removePosts, _removeComments, _title, _flairTemplateID)


@dataclass
class RemovalReasons:
	pmsubject: str
	logreason: str
	header: str
	footer: str
	removalOption: str
	typeReply: str
	typeStickied: bool
	typeCommentAsSubreddit: bool
	typeLockComment: bool
	typeAsSub: bool
	autoArchive: bool
	typeLockThread: bool
	logsub: str
	logtitle: str
	bantitle: str
	getfrom: str
	reasons: List[Reason]

	def to_dict(RemovalReasons):
		d = {
		"pmsubject": urllib.parse.quote(RemovalReasons.pmsubject),
		"logreason": urllib.parse.quote(RemovalReasons.logreason),
		"header": urllib.parse.quote(RemovalReasons.header),
		"footer": urllib.parse.quote(RemovalReasons.footer),
		"removalOption": RemovalReasons.removalOption,
		"typeReply": RemovalReasons.typeReply,
		"typeStickied": RemovalReasons.typeStickied,
		"typeCommentAsSubreddit": RemovalReasons.typeCommentAsSubreddit,
		"typeLockComment": RemovalReasons.typeLockComment,
		"typeAsSub": RemovalReasons.typeAsSub,
		"autoArchive": RemovalReasons.autoArchive,
		"typeLockThread": RemovalReasons.typeLockThread,
		"logsub": urllib.parse.quote(RemovalReasons.logsub),
		"logtitle":RemovalReasons.logtitle,
		"bantitle": urllib.parse.quote(RemovalReasons.bantitle),
		"getfrom": RemovalReasons.getfrom,
		"reasons":[Reason.to_dict(r) for r in RemovalReasons.reasons]
		}
		return d

	@staticmethod
	def from_dict(obj: Any) -> 'RemovalReasons':
		_pmsubject = urllib.parse.unquote(str(obj.get("pmsubject")))
		_logreason = urllib.parse.unquote(str(obj.get("logreason")))
		_header = urllib.parse.unquote(str(obj.get("header")))
		_footer = urllib.parse.unquote(str(obj.get("footer")))
		_removalOption = str(obj.get("removalOption"))
		_typeReply = str(obj.get("typeReply"))
		_typeStickied = bool(obj.get("typeStickied"))
		_typeCommentAsSubreddit = bool(obj.get("typeCommentAsSubreddit"))
		_typeLockComment = bool(obj.get("typeLockComment"))
		_typeAsSub = bool(obj.get("typeAsSub"))
		_autoArchive = bool(obj.get("autoArchive"))
		_typeLockThread = bool(obj.get("typeLockThread"))
		_logsub = urllib.parse.unquote(str(obj.get("logsub")))
		_logtitle = str(obj.get("logtitle"))
		_bantitle = urllib.parse.unquote(str(obj.get("bantitle")))
		_getfrom = str(obj.get("getfrom"))
		_reasons = [Reason.from_dict(y) for y in obj.get("reasons")]
		return RemovalReasons(_pmsubject, _logreason, _header, _footer, _removalOption, _typeReply, _typeStickied, _typeCommentAsSubreddit, _typeLockComment, _typeAsSub, _autoArchive, _typeLockThread, _logsub, _logtitle, _bantitle, _getfrom, _reasons)


@dataclass
class UsernoteColor:
	key: str
	text: str
	color: str

	@staticmethod
	def from_dict(obj: Any) -> 'UsernoteColor':
		_key = str(obj.get("key"))
		_text = str(obj.get("text"))
		_color = str(obj.get("color"))
		return UsernoteColor(_key, _text, _color)


@dataclass
class SettingsRoot:
	ver: int
	domainTags: List[DomainTag]
	removalReasons: RemovalReasons
	modMacros: List[ModMacro]
	usernoteColors: List[UsernoteColor]
	banMacros: BanMacros

	def __repr__(self):
		return f"SettingsRoot(ver='{self.ver}')"

	def __str__(self):
		return self.to_json()

	def to_json(self):
		return json.dumps(self.__dict__, cls=JSONEncoder)

	@staticmethod
	def from_dict(obj: Any) -> 'SettingsRoot':
		_ver = int(obj.get("ver"))
		_domainTags = [DomainTag.from_dict(y) for y in obj.get("domainTags")]
		_removalReasons = RemovalReasons.from_dict(obj.get("removalReasons"))
		_modMacros = [ModMacro.from_dict(y) for y in obj.get("modMacros")]
		_usernoteColors = [UsernoteColor.from_dict(y) for y in obj.get("usernoteColors")]
		_banMacros = BanMacros.from_dict(obj.get("banMacros"))
		return SettingsRoot(_ver, _domainTags, _removalReasons, _modMacros, _usernoteColors, _banMacros)


class ToolboxSettings:
	"""Represents the Toolbox Settings page."""
	def __init__(self, subreddit, identifier=DEFAULT_IDENTIFIER, lazy=False):
		"""
		Construtor for the ToolboxSettings class.

		Parameters
		----------
		subreddit: praw.subreddit object
			the subreddit to use the Toolbox object with
		lazy: Bool
			if set to True, does not fetch from Reddit on instantiation
		identifier: String
			string that get's appended to all wikiedit discriptions identifying
			pmtw as the actioner. Defaults to "via pmtw"
		"""
		
		self.__subreddit = subreddit
		self.identifier = identifier
		self.__settings = ""
		self.ver = ""
		self.domainTags = ""
		self.removalReasons = ""
		self.modMacros = ""
		self.usernoteColors = ""
		self.banMacros = ""
		self.warnings = ""

		if not lazy: self.load()

	def __repr__(self):
		"""Set display for a ToolboxSettings object"""
		return f"ToolboxSettings(subreddit='{self.__subreddit}')"

	def load(self):
		"""
		Load Toolbox Settings from Reddit

		Returns
		-------
		String
			Information letting the user know settings are loaded
		"""

		try:
			page = self.__subreddit.wiki[SETTINGS_PAGE].content_md
			page = json.loads(page)
			if page["ver"] != 1: raise ValueError(f"pmtw requires settings ver {SETTINGS_VERSION}, got {page['ver']}")
		except NotFound:
			initialJSON = {"ver":1,"domainTags":"","removalReasons":{"pmsubject":"","logreason":"","header":"test","footer":"","removalOption":"suggest","typeReply":"reply","typeStickied":False,"typeCommentAsSubreddit":False,"typeLockComment":False,"typeAsSub":False,"autoArchive":False,"typeLockThread":False,"logsub":"","logtitle":"","bantitle":"","getfrom":"","reasons":[]},"modMacros":[],"usernoteColors":[{"key":"gooduser","text":"Good Contributor","color":"#008000"},{"key":"spamwatch","text":"Spam Watch","color":"#ff00ff"},{"key":"spamwarn","text":"Spam Warning","color":"#800080"},{"key":"abusewarn","text":"Abuse Warning","color":"#ffa500"},{"key":"ban","text":"Ban","color":"#ff0000"},{"key":"permban","text":"Permanent Ban","color":"#8b0000"},{"key":"botban","text":"Bot Ban","color":"#000000"}],"banMacros":{"banNote":"","banMessage":""}}
			page = self.__subreddit.wiki.create(name=SETTINGS_PAGE, content=str(initialJSON))
			self.__subreddit.wiki[SETTINGS_PAGE].mod.update(listed=False, permlevel=2)

		# append expected items to the json, in case the toolbox settings 
		# don't have them, since things get added here without a bump to ver
		for item in ["domainTags", "removalReasons", "modMacros","banMacros"]:
			if item not in page.keys(): page[item] = ""

		# Copy things over so we don't have to hit ToolboxSettings.settings
		# for every variable
		self.__settings = SettingsRoot.from_dict(page)
		self.ver = self.__settings.ver
		self.domainTags = self.__settings.domainTags
		self.removalReasons = self.__settings.removalReasons
		self.modMacros = self.__settings.modMacros
		self.usernoteColors = self.__settings.usernoteColors
		self.banMacros = self.__settings.banMacros

		warnings = []
		for color in self.usernoteColors:
			warnings.append(color.key)
		self.warnings = warnings

		return "Settings loaded"

	def save(self, reason="Settings update"):
		"""
		Save Toolbox settings back to Reddit

		Parameters
		----------
		reason: String
			Optional, set a custom reason for the wiki page description

		Returns
		-------
		String
			Reason with identifier set when instantiating a Toolbox object

		Raises
		------
		OverflowError
			If the settings page is too big to save to Reddit

		"""
		reason = f"{reason} {self.identifier}"

		# Copy the class variables back to the dataclasses for compression
		self.__settings.ver = self.ver
		self.__settings.domainTags = self.domainTags
		self.__settings.removalReasons = self.removalReasons
		self.__settings.modMacros = self.modMacros
		self.__settings.usernoteColors = self.usernoteColors
		self.__settings.banMacros = self.banMacros

		if len(self.__settings.__str__()) > MAX_WIKI_SIZE:
			raise OverflowError(f'Usernote data {len(self.__settings.__str__()) - MAX_WIKI_SIZE} bytes too big to insert')
		self.__subreddit.wiki[SETTINGS_PAGE].edit(content=self.__settings.__str__(),reason=reason)
		return reason

	def stream(self, pause_after=None, skip_existing=False):
		"""
		Yields new WikiRevision objects as they become available

		Parameters
		----------
		pause_after: [Optional] Integer (Default: `None`)
			An integer representing the number of requests that result in no new 
			items before this function yields `None`, effectively introducing a 
			pause into the stream. A negative value yields `None` after items 
			from a single response have been yielded, regardless of number of new 
			items obtained in that response. A value of `0` yields `None` after 
			every response resulting in no new items, and a value of `None` never 
			introduces a pause.
		skip_existing: [Optional] Boolean (Default: False)
			When `True`, this does not yield any results from the first request 
			thereby skipping any items that existed in the stream prior to starting 
			the stream.

		Yields
		------
		WikiRevision objects
		
		"""
		for revision in revisions_stream(
			sub=self.__subreddit,
			page=SETTINGS_PAGE,
			pause_after = pause_after,
			skip_existing = skip_existing
		):
			yield revision



import json
import time as t

from pmtw.usernotes import ToolboxNote, ToolboxUsernotes

"""
Wrappers to provide functionality for users coming from puni
"""

class puni_Note:
	warnings = ['none','spamwatch','spamwarn','abusewarn','ban','permban','botban','gooduser']

	def __init__(self, user, note, subreddit=None, mod=None, link='', warning='none', note_time=int(t.time())):
		self.__n = ToolboxNote(user, note, time=note_time, mod=mod, warning=warning, url=link)
		
		self.username = self.__n.user
		self.note = self.__n.note
		self.subreddit = str(subreddit) if subreddit else None
		self.time = self.__n.time
		self.moderator = self.__n.mod
		self.link = self.__n.link

	def __str__(self):
		return f'{self.username}: {self.note}'

	def __repr__(self):
		return f"Note(username='{self.username}', ver='puni')"

	def full_url(self):
		return self.__n.url

	@staticmethod
	def _compress_url(link):
		return ToolboxNote._ToolboxNote__compress_url(link)

	@staticmethod
	def _expand_url(short_link, subreddit=None):
		if not subreddit: raise ValueError('Subreddit name must be provided')
		return ToolboxNote._ToolboxNote__expand_link(short_link)


class puni_UserNotes:

	schema = 6
	max_page_size = 524288
	zlib_compression_strength = 9
	page_name = 'usernotes'

	def __init__(self, r, subreddit, lazy_start=False):
		self.r = subreddit._reddit
		self.subreddit = subreddit
		self.cached_json = {}
		self.__ToolboxUsernotes = ToolboxUsernotes(self.subreddit, identifier="via puni", settingsWarnings=puni_Note.warnings)

		if not lazy_start: self.get_json()

	def __repr__(self):
		return f"UserNotes(subreddit='{self.subreddit.display_name}')"

	def get_json(self):
		self.__ToolboxUsernotes.load()
		self.cached_json = self.__ToolboxUsernotes._ToolboxUsernotes__usernotesJSON

	def _init_notes(self):
		self.get_json()

	def set_json(self, reason=None, new_page=False):
		if new_page: self.get_json()
		self.__ToolboxUsernotes.save(reason)

	def get_notes(self, user, lazy=False):
		if lazy == True: fresh = False
		else: fresh = True
		return self.__ToolboxUsernotes.list_notes(user, fresh=fresh)

	def get_users(self, lazy=False):
		if lazy == True: fresh = False
		else: fresh = True
		return self.__ToolboxUsernotes.list_users(fresh=fresh)

	def _mod_from_index(self, index):
		return self.__ToolboxUsernotes._ToolboxUsernotes__usernotesJSON['constants']['users'][index]

	def _warning_from_index(self, index):
		return self.__ToolboxUsernotes._ToolboxUsernotes__usernotesJSON['constants']['warnings'][index]

	def _expand_json(self, j):
		return self.__ToolboxUsernotes._ToolboxUsernotes__expand_json(j)

	def _compress_json(self, j):
		return json.loads(self.__ToolboxUsernotes._ToolboxUsernotes__compress_json(j))

	def add_note(self, note, lazy=False):
		self.__ToolboxUsernotes.add(note, lazy)
		self.cached_json = self.__ToolboxUsernotes._ToolboxUsernotes__usernotesJSON
		return f'"create new note on user {note.user}" via puni'

	def remove_note(self, user, index, lazy=False):
		if lazy == True: fresh = False
		else: fresh = True
		idx = 0
		popped = False
		for note in self.__ToolboxUsernotes.list_notes(user, fresh=fresh):
			if idx == index:
				self.__ToolboxUsernotes.remove(user, note.time)
				popped = True
			idx += 1
		if popped:
			self.cached_json = self.__ToolboxUsernotes._ToolboxUsernotes__usernotesJSON
			return f'"delete note #{index} on user {user}" via puni'
		else:
			raise IndexError("pop index out of range")

	def remove_user(self, user, lazy=False):
		if lazy == True: fresh = False
		else: fresh = True
		self.__ToolboxUsernotes.remove(user, fresh=fresh)
		self.cached_json = self.__ToolboxUsernotes._ToolboxUsernotes__usernotesJSON
		return f'"delete user {user} from usernotes" via puni'



import json
import urllib.parse

from pmtw.usernotes import ToolboxNote, ToolboxUsernotes
from pmtw.toolbox import Toolbox

"""
Wrappers to provide drop-in functionality for users coming from pmtw version 0.2.1
"""

class Note:
	"""
	Wrapper for ToolboxNote, to act as a drop-in replacement for 
	pmtw version 0.2.1
	"""
	def __init__(self, user, note, time=None, mod=None, warning=None, link='', subreddit=None):
		self.__n = ToolboxNote(user, note, time=time, mod=mod, warning=warning, url=link)
		self.user = self.__n.user
		self.time = self.__n.time
		self.note = self.__n.note
		self.mod = self.__n.mod
		self.warning = self.__n.warning
		self.subreddit = subreddit

		if "l," in self.__n.link:
			self.link = f"https://redd.it/{self.__n.link.split(',')[1]}"
		else:
			self.link = link

	def __repr__(self):
		return f"Note(user='{self.user}', note='{self.note}', time='{self.time}')"


class Usernotes:
	"""
	Wrapper for ToolboxUsernotes, to act as a drop-in replacement for 
	pmtw version 0.2.1
	"""
	def __init__(self, r, subreddit, fetch=True):
		self.r = subreddit._reddit
		self.subreddit = subreddit
		self.usernotesJSON = {}
		self.__ToolboxUsernotes = ToolboxUsernotes(self.subreddit, identifier="via pmtw")

		if fetch == True:
			self.__ToolboxUsernotes.load()
			self.usernotesJSON = self.__ToolboxUsernotes._ToolboxUsernotes__usernotesJSON

	def __repr__(self):
		return f"Usernotes(subreddit='{self.subreddit}')"

	def __mod_index(self, mod, type="decode"):
		if type == "decode":
			return self.usernotesJSON['constants']['users'][mod]
		else:
			return self.__ToolboxUsernotes._ToolboxUsernotes__get_mod_index(mod)

	def __warnings_index(self, warning, type ="decode"):
		if type == "decode":
			return self.__ToolboxUsernotes.warnings.index(warning)
		else:
			return self.__ToolboxUsernotes._ToolboxUsernotes__get_warning_index(warning)

	def __expand_json(self, notes):
		return self.__ToolboxUsernotes._ToolboxUsernotes__expand_json(notes)
	
	def __compress_json(self, notes):
		return json.loads(self.__ToolboxUsernotes._ToolboxUsernotes__compress_json(notes))

	def __fetch_from_reddit(self):
		self.__ToolboxUsernotes.load()

	def __expand_link(self, link):
		return ToolboxNote._ToolboxNote__expand_link(link)

	def __compress_link(self, link):
		print(link)
		l = ToolboxNote._ToolboxNote__compress_url(link)
		return l

	def push_usernotes(self, reason='"Batch usernote update"', truncate=False):
		self.__ToolboxUsernotes.save(reason=reason)
	
	def warnings_types(self, local=True):
		if local == False: self.__ToolboxUsernotes.load()
		return self.__ToolboxUsernotes.warnings

	def get_user_notes(self, user, local=True):
		if local == True: fresh = False
		else: fresh = True
		return self.__ToolboxUsernotes.list_notes(user,fresh=fresh)

	def existing_users(self, local=True):
		if local == True: fresh = False
		else: fresh = True
		return self.__ToolboxUsernotes.list_users(fresh=fresh)

	def add_note(self, note, batch=False):
		print(note._Note__n.url) #preserve quirk with version 0.2.1
		if note.warning not in self.__ToolboxUsernotes.warnings:
			raise ValueError(f"{note.warning} is not a valid warning kind.")
		return self.__ToolboxUsernotes.add(note._Note__n, bulk_action=batch)

	def delete_note(self, username, timestamp, batch=False):
		return self.__ToolboxUsernotes.remove(username, timestamp, bulk_action=batch)

	def purge_user_notes(self, username, batch=False):
		return self.__ToolboxUsernotes.remove(username, bulk_action=batch)


class Settings:
	"""
	Wrapper for ToolboxSettings, to act as a drop-in replacement for 
	pmtw version 0.2.1
	"""
	def __init__(self, r, subreddit, fetch=True):
		self.r = subreddit._reddit
		self.subreddit = subreddit
		self.settingsJSON = {}
		self.__Toolbox = Toolbox(subreddit)

		if fetch:
			self.__Toolbox.load_settings()
			self.settingsJSON = self.__Toolbox.settings

	def __repr__(self):
		return f"ToolboxSettings(subreddit='{self.subreddit}')"

	def __decode_text(self,text):
		return urllib.parse.unquote(text)

	def __encode_text(self, text):
		return urllib.parse.quote(text)
	
	def __fetch_from_reddit(self):
		self.__Toolbox.load_settings()

	def push_settings(self, reason='Toolbox Settings update from pmtw'):
		self.__Toolbox.save_settings()

	def test(self):
		return self.settingsJSON

	def get_domainTags(self, local=True):
		if local == False: self.__Toolbox.load_settings()
		return self.__Toolbox.settings.domainTags

	def get_removalReasons(self, local=True):
		if local == False: self.__Toolbox.load_settings()
		return self.__Toolbox.settings.removalReasons

	def get_reasons(self, local=True):
		if local == False: self.__Toolbox.load_settings()
		self.__Toolbox.settings.removalReasons.reasons

	def add_reason(self, reason, local=True):
		if local == False: self.__Toolbox.load_settings()
		self.__Toolbox.settings.removalReasons.reasons.append(reason)

	def get_modMacros(self, local=True):
		if local == False: self.__Toolbox.load_settings()
		return self.__Toolbox.settings.modMacros

	def get_usernoteColors(self, local=True):
		if local == False: self.__Toolbox.load_settings()
		return self.__Toolbox.settings.usernoteColors

	def get_banMacros(self, local=True):
		if local == False: self.__Toolbox.load_settings()
		return self.__Toolbox.settings.banMacros



import time
from typing import Any, Generator, Optional
from praw.models.util import BoundedSet, ExponentialCounter
from praw.models import ListingGenerator
from datetime import datetime

class WikiRevision:
	def __init__(self, user, timestamp, page, revision_hidden, reason, id):
		self.user = user
		self.timestamp = timestamp
		self.page = page
		self.revision_hidden = revision_hidden
		self.reason = reason
		self.id = id
		
		self.human_time = datetime.fromtimestamp(self.timestamp).__str__()

	def __repr__(self):
		return f"WikiRevision(page='{self.page}' user='{self.user}' human_time='{self.human_time}')"

def revisions_stream(
	sub=None,
	page=None,
	attribute_name: str = "id",
	pause_after: Optional[int] = None,
	skip_existing: bool = False,
	**function_kwargs: Any,
) -> Generator[Any, None, None]:
	"""
	Yield new items from a wikipage of a subreddit as they become available.

	modified from https://github.com/praw-dev/praw/blob/master/praw/models/util.py

	Parameters
	----------
	sub: praw.subreddit object
		The subreddit to stream wiki revisions from
	page: String
		the wiki page to stream from. If it's a subpage, you should pass the 
		path in the format `<folder>/<page>`
	pause_after: [Optional] Integer (Default: `None`)
		An integer representing the number of requests that result in no new 
		items before this function yields `None`, effectively introducing a 
		pause into the stream. A negative value yields `None` after items 
		from a single response have been yielded, regardless of number of new 
		items obtained in that response. A value of `0` yields `None` after 
		every response resulting in no new items, and a value of `None` never 
		introduces a pause.
	skip_existing: [Optional] Boolean (Default: False)
		When `True`, this does not yield any results from the first request 
		thereby skipping any items that existed in the stream prior to starting 
		the stream.

	additonal keyword arguments are passed to the inner function

	Yields
	------
	WikiRevision object

	Note
	----
	This function internally uses an exponential delay with jitter between 
	subsequent responses that contain no new results, up to a maximum delay of 
	just over 16 seconds. In practice, that means that the time before pause 
	for `pause_after=N+1` is approximately twice the time before pause for 
	`pause_after=N`.


	"""
	path = f'/r/{sub.display_name}/wiki/revisions/{page}'
	def function(limit=None, params=None):
		if params:
			params["before"] = "WikiRevision_" + params["before"]
		return ListingGenerator(
			reddit=sub._reddit,
			url=path,
			params=params
		)

	before_attribute = None
	exponential_counter = ExponentialCounter(max_counter=16)
	seen_attributes = BoundedSet(301)
	without_before_counter = 0
	responses_without_new = 0
	valid_pause_after = pause_after is not None
	while True:
		found = False
		newest_attribute = None
		limit = 100
		if before_attribute is None:
			limit -= without_before_counter
			without_before_counter = (without_before_counter + 1) % 30
		for item in reversed(list(function(limit=limit, **function_kwargs))):
			item = WikiRevision(
				sub._reddit.redditor(item["author"]["data"]["name"]),
				item["timestamp"],
				sub.wiki[item["page"]].revision(item["id"]),
				item["revision_hidden"],
				item["reason"],
				item["id"]
			)
			attribute = getattr(item, attribute_name)
			if attribute in seen_attributes: continue
			found = True
			seen_attributes.add(attribute)
			newest_attribute = attribute
			if not skip_existing: 
				yield item
		before_attribute = newest_attribute
		skip_existing = False
		if valid_pause_after and pause_after < 0:
			yield None
		elif found:
			exponential_counter.reset()
			responses_without_new = 0
		else:
			responses_without_new += 1
			if valid_pause_after and responses_without_new > pause_after:
				exponential_counter.reset()
				responses_without_new = 0
				yield None
			else:
				time.sleep(exponential_counter.counter())




#!/usr/bin/env python
"""pmtw setup.py"""

import setuptools

if __name__ == "__main__":
    setuptools.setup()

"""
from setuptools import setup
from os import path
import re
from codecs import open

PACKAGE_NAME = "pmtw"
HERE = path.abspath(path.dirname(__file__))
with open(path.join(HERE, "README.md"), encoding="utf-8") as fp:
    README = fp.read()
with open(path.join(HERE, PACKAGE_NAME, "constants.py"), encoding="utf-8") as fp:
    VERSION = re.search('__version__ = "([^"]+)"', fp.read()).group(1)

setup(
  name = PACKAGE_NAME,
  author = 'adhesiveCheese',
  author_email = 'adhesiveCheese@gmail.com',
  python_requires="~=3.7",
  classifiers=[
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Utilities",
  ],
  description = (
        'PMTW, an acronym for "Python Moderator Toolbox Wrapper", is a python '
        "package that allows for simple access to Moderator Toolbox"
  ),
  install_requires=[
          'praw >=7.0',
  ],
  keywords = ['Reddit', 'Moderator_Toolbox', 'Web Wrapper'],
  license="Simplified BSD License",
  url = 'https://github.com/adhesivecheese/pmtw',
  download_url = 'https://github.com/adhesivecheese/pmtw/archive/refs/tags/v1_0_0.tar.gz',
  version = VERSION,
  long_description = README,

)
"""